@@ -1,17 +1,17 @@
-import time
-import pickle
+import hashlib
+import inspect
 import logging
 import os
-import hashlib
-from shutil import move, rmtree
+import pickle
 import tempfile
-import inspect
+import time
+from shutil import move, rmtree
 
 from fsspec import AbstractFileSystem, filesystem
-from fsspec.spec import AbstractBufferedFile
-from fsspec.core import MMapCache, BaseCache
-from fsspec.utils import infer_compression
 from fsspec.compression import compr
+from fsspec.core import BaseCache, MMapCache
+from fsspec.spec import AbstractBufferedFile
+from fsspec.utils import infer_compression
 
 logger = logging.getLogger("fsspec")
 
@@ -48,7 +48,7 @@ class CachingFileSystem(AbstractFileSystem):
         fs=None,
         same_names=False,
         compression=None,
-        **kwargs
+        **kwargs,
     ):
         """
 
@@ -105,6 +105,8 @@ class CachingFileSystem(AbstractFileSystem):
         self.check_files = check_files
         self.expiry = expiry_time
         self.compression = compression
+        # TODO: same_names should allow for variable prefix, not only
+        #  to keep the basename
         self.same_names = same_names
         self.target_protocol = (
             target_protocol
@@ -152,11 +154,12 @@ class CachingFileSystem(AbstractFileSystem):
             with open(fn, "rb") as f:
                 cached_files = pickle.load(f)
             for k, c in cached_files.items():
-                if c["blocks"] is not True:
-                    if cache[k]["blocks"] is True:
+                if k in cache:
+                    if c["blocks"] is True or cache[k]["blocks"] is True:
                         c["blocks"] = True
                     else:
                         c["blocks"] = set(c["blocks"]).union(cache[k]["blocks"])
+                    c["time"] = max(c["time"], cache[k]["time"])
 
             # Files can be added to cache after it was written once
             for k, c in cache.items():
@@ -173,6 +176,8 @@ class CachingFileSystem(AbstractFileSystem):
             pickle.dump(cache, f)
         self._mkcache()
         move(fn2, fn)
+        self.cached_files[-1] = cached_files
+        self.last_cache = time.time()
 
     def _check_cache(self):
         """Reload caches if time elapsed or any disappeared"""
@@ -190,20 +195,15 @@ class CachingFileSystem(AbstractFileSystem):
         path = self._strip_protocol(path)
 
         self._check_cache()
-        if not path.startswith(self.target_protocol):
-            store_path = self.target_protocol + "://" + path
-            path = self.fs._strip_protocol(store_path)
-        else:
-            store_path = path
         for storage, cache in zip(self.storage, self.cached_files):
-            if store_path not in cache:
+            if path not in cache:
                 continue
-            detail = cache[store_path].copy()
+            detail = cache[path].copy()
             if self.check_files:
                 if detail["uid"] != self.fs.ukey(path):
                     continue
             if self.expiry:
-                if detail["time"] - time.time() > self.expiry:
+                if time.time() - detail["time"] > self.expiry:
                     continue
             fn = os.path.join(storage, detail["fn"])
             if os.path.exists(fn):
@@ -227,18 +227,13 @@ class CachingFileSystem(AbstractFileSystem):
         raises PermissionError
         """
         path = self._strip_protocol(path)
-        if not path.startswith(self.target_protocol):
-            store_path = self.target_protocol + "://" + path
-            path = self.fs._strip_protocol(store_path)
-        else:
-            store_path = path
         _, fn = self._check_file(path)
         if fn is None:
             return
         if fn.startswith(self.storage[-1]):
             # is in in writable cache
             os.remove(fn)
-            self.cached_files[-1].pop(store_path)
+            self.cached_files[-1].pop(path)
             self.save_cache()
         else:
             raise PermissionError(
@@ -252,7 +247,7 @@ class CachingFileSystem(AbstractFileSystem):
         block_size=None,
         autocommit=True,
         cache_options=None,
-        **kwargs
+        **kwargs,
     ):
         """Wrap the target _open
 
@@ -268,11 +263,7 @@ class CachingFileSystem(AbstractFileSystem):
         """
         path = self._strip_protocol(path)
 
-        if not path.startswith(self.target_protocol):
-            store_path = self.target_protocol + "://" + path
-        else:
-            store_path = path
-        path = self.fs._strip_protocol(store_path)
+        path = self.fs._strip_protocol(path)
         if "r" not in mode:
             return self.fs._open(
                 path,
@@ -280,9 +271,9 @@ class CachingFileSystem(AbstractFileSystem):
                 block_size=block_size,
                 autocommit=autocommit,
                 cache_options=cache_options,
-                **kwargs
+                **kwargs,
             )
-        detail = self._check_file(store_path)
+        detail = self._check_file(path)
         if detail:
             # file is in cache
             detail, fn = detail
@@ -294,16 +285,17 @@ class CachingFileSystem(AbstractFileSystem):
             # TODO: action where partial file exists in read-only cache
             logger.debug("Opening partially cached copy of %s" % path)
         else:
-            hash = hash_name(path, self.same_names)
+            hash = self.hash_name(path, self.same_names)
             fn = os.path.join(self.storage[-1], hash)
             blocks = set()
             detail = {
+                "original": path,
                 "fn": hash,
                 "blocks": blocks,
                 "time": time.time(),
                 "uid": self.fs.ukey(path),
             }
-            self.cached_files[-1][store_path] = detail
+            self.cached_files[-1][path] = detail
             logger.debug("Creating local sparse file for %s" % path)
 
         # call target filesystems open
@@ -315,7 +307,7 @@ class CachingFileSystem(AbstractFileSystem):
             autocommit=autocommit,
             cache_options=cache_options,
             cache_type=None,
-            **kwargs
+            **kwargs,
         )
         if self.compression:
             comp = (
@@ -339,17 +331,28 @@ class CachingFileSystem(AbstractFileSystem):
         self.save_cache()
         return f
 
+    def hash_name(self, path, same_name):
+        return hash_name(path, same_name=same_name)
+
     def close_and_update(self, f, close):
         """Called when a file is closing, so store the set of blocks"""
+        if f.closed:
+            return
         path = self._strip_protocol(f.path)
 
-        if not path.startswith(self.target_protocol):
-            store_path = self.target_protocol + "://" + path
-        c = self.cached_files[-1][store_path]
+        c = self.cached_files[-1][path]
         if c["blocks"] is not True and len(["blocks"]) * f.blocksize >= f.size:
             c["blocks"] = True
-        self.save_cache()
+        try:
+            logger.debug("going to save")
+            self.save_cache()
+            logger.debug("saved")
+        except (IOError, OSError):
+            logger.debug("Cache saving failed while closing file")
+        except NameError:
+            logger.debug("Cache save failed due to interpreter shutdown")
         close()
+        f.closed = True
 
     def __getattribute__(self, item):
         if item in [
@@ -360,6 +363,7 @@ class CachingFileSystem(AbstractFileSystem):
             "__init__",
             "__getattribute__",
             "__reduce__",
+            "_make_local_details",
             "open",
             "cat",
             "cat_file",
@@ -369,17 +373,25 @@ class CachingFileSystem(AbstractFileSystem):
             "head",
             "_check_file",
             "_check_cache",
+            "_mkcache",
             "clear_cache",
             "pop_from_cache",
             "_mkcache",
             "local_file",
             "_paths_from_path",
+            "get_mapper",
             "open_many",
             "commit_many",
+            "hash_name",
+            "__hash__",
+            "__eq__",
+            "to_json",
         ]:
             # all the methods defined in this class. Note `open` here, since
             # it calls `_open`, but is actually in superclass
-            return lambda *args, **kw: getattr(type(self), item)(self, *args, **kw)
+            return lambda *args, **kw: getattr(type(self), item).__get__(self)(
+                *args, **kw
+            )
         if item in ["__reduce_ex__"]:
             raise AttributeError
         if item in ["_cache"]:
@@ -398,7 +410,7 @@ class CachingFileSystem(AbstractFileSystem):
             # attributed belonging to the target filesystem
             cls = type(fs)
             m = getattr(cls, item)
-            if inspect.isfunction(m) and (
+            if (inspect.isfunction(m) or inspect.isdatadescriptor(m)) and (
                 not hasattr(m, "__self__") or m.__self__ is None
             ):
                 # instance method
@@ -408,6 +420,45 @@ class CachingFileSystem(AbstractFileSystem):
             # attributes of the superclass, while target is being set up
             return super().__getattribute__(item)
 
+    def __eq__(self, other):
+        """Test for equality."""
+        if self is other:
+            return True
+        if not isinstance(other, type(self)):
+            return False
+        return (
+            self.storage == other.storage
+            and self.kwargs == other.kwargs
+            and self.cache_check == other.cache_check
+            and self.check_files == other.check_files
+            and self.expiry == other.expiry
+            and self.compression == other.compression
+            and self.same_names == other.same_names
+            and self.target_protocol == other.target_protocol
+        )
+
+    def __hash__(self):
+        """Calculate hash."""
+        return (
+            hash(tuple(self.storage))
+            ^ hash(str(self.kwargs))
+            ^ hash(self.cache_check)
+            ^ hash(self.check_files)
+            ^ hash(self.expiry)
+            ^ hash(self.compression)
+            ^ hash(self.same_names)
+            ^ hash(self.target_protocol)
+        )
+
+    def to_json(self):
+        """Calculate JSON representation.
+
+        Not implemented yet for CachingFileSystem.
+        """
+        raise NotImplementedError(
+            "CachingFileSystem JSON representation not implemented"
+        )
+
 
 class WholeFileCacheFileSystem(CachingFileSystem):
     """Caches whole remote files on first access
@@ -427,7 +478,7 @@ class WholeFileCacheFileSystem(CachingFileSystem):
     local_file = True
 
     def open_many(self, open_files):
-        paths, store_paths = zip(*[self._paths_from_path(of.path) for of in open_files])
+        paths = [of.path for of in open_files]
         if "r" in open_files.mode:
             self._mkcache()
         else:
@@ -438,11 +489,10 @@ class WholeFileCacheFileSystem(CachingFileSystem):
 
         if self.compression:
             raise NotImplementedError
-        details = [self._check_file(sp) for sp in store_paths]
+        details = [self._check_file(sp) for sp in paths]
         downpath = [p for p, d in zip(paths, details) if not d]
-        downstore = [p for p, d in zip(store_paths, details) if not d]
         downfn0 = [
-            os.path.join(self.storage[-1], hash_name(p, self.same_names))
+            os.path.join(self.storage[-1], self.hash_name(p, self.same_names))
             for p, d in zip(paths, details)
         ]  # keep these path names for opening later
         downfn = [fn for fn, d in zip(downfn0, details) if not d]
@@ -453,7 +503,8 @@ class WholeFileCacheFileSystem(CachingFileSystem):
             # update metadata - only happens when downloads are successful
             newdetail = [
                 {
-                    "fn": hash_name(path, self.same_names),
+                    "original": path,
+                    "fn": self.hash_name(path, self.same_names),
                     "blocks": True,
                     "time": time.time(),
                     "uid": self.fs.ukey(path),
@@ -461,7 +512,7 @@ class WholeFileCacheFileSystem(CachingFileSystem):
                 for path in downpath
             ]
             self.cached_files[-1].update(
-                {store_path: detail for store_path, detail in zip(downstore, newdetail)}
+                {path: detail for path, detail in zip(downpath, newdetail)}
             )
             self.save_cache()
 
@@ -477,70 +528,92 @@ class WholeFileCacheFileSystem(CachingFileSystem):
     def commit_many(self, open_files):
         self.fs.put([f.fn for f in open_files], [f.path for f in open_files])
 
-    def _paths_from_path(self, path):
-        path = self._strip_protocol(path)
+    def _make_local_details(self, path):
+        hash = self.hash_name(path, self.same_names)
+        fn = os.path.join(self.storage[-1], hash)
+        detail = {
+            "original": path,
+            "fn": hash,
+            "blocks": True,
+            "time": time.time(),
+            "uid": self.fs.ukey(path),
+        }
+        self.cached_files[-1][path] = detail
+        logger.debug("Copying %s to local cache" % path)
+        return fn
 
-        if not path.startswith(self.target_protocol):
-            store_path = self.target_protocol + "://" + path
-        else:
-            store_path = path
-        path = self.fs._strip_protocol(store_path)
-        return path, store_path
+    def cat(self, path, recursive=False, on_error="raise", **kwargs):
+        paths = self.expand_path(
+            path, recursive=recursive, maxdepth=kwargs.get("maxdepth", None)
+        )
+        getpaths = []
+        storepaths = []
+        fns = []
+        for p in paths:
+            detail = self._check_file(p)
+            if not detail:
+                fn = self._make_local_details(p)
+                getpaths.append(p)
+                storepaths.append(fn)
+            else:
+                detail, fn = detail if isinstance(detail, tuple) else (None, detail)
+            fns.append(fn)
+        if getpaths:
+            self.fs.get(getpaths, storepaths)
+            self.save_cache()
+        out = {path: open(fn, "rb").read() for path, fn in zip(paths, fns)}
+        if isinstance(path, str) and len(paths) == 1 and recursive is False:
+            out = out[paths[0]]
+        return out
 
     def _open(self, path, mode="rb", **kwargs):
-        path, store_path = self._paths_from_path(path)
+        path = self._strip_protocol(path)
         if "r" not in mode:
             return self.fs._open(path, mode=mode, **kwargs)
-        detail = self._check_file(store_path)
+        detail = self._check_file(path)
         if detail:
             detail, fn = detail
-            hash, blocks = detail["fn"], detail["blocks"]
+            _, blocks = detail["fn"], detail["blocks"]
             if blocks is True:
                 logger.debug("Opening local copy of %s" % path)
-                return open(fn, mode)
+
+                # In order to support downstream filesystems to be able to
+                # infer the compression from the original filename, like
+                # the `TarFileSystem`, let's extend the `io.BufferedReader`
+                # fileobject protocol by adding a dedicated attribute
+                # `original`.
+                f = open(fn, mode)
+                f.original = detail.get("original")
+                return f
             else:
                 raise ValueError(
                     "Attempt to open partially cached file %s"
                     "as a wholly cached file" % path
                 )
         else:
-            hash = hash_name(path, self.same_names)
-            fn = os.path.join(self.storage[-1], hash)
-            detail = {
-                "fn": hash,
-                "blocks": True,
-                "time": time.time(),
-                "uid": self.fs.ukey(path),
-            }
-            self.cached_files[-1][store_path] = detail
-            logger.debug("Copying %s to local cache" % path)
+            fn = self._make_local_details(path)
         kwargs["mode"] = mode
 
         # call target filesystems open
-        # TODO: why not just use fs.get ??
-        f = self.fs._open(path, **kwargs)
-        if self.compression:
-            comp = (
-                infer_compression(path)
-                if self.compression == "infer"
-                else self.compression
-            )
-            f = compr[comp](f, mode="rb")
         self._mkcache()
-        with open(fn, "wb") as f2:
-            if isinstance(f, AbstractBufferedFile):
-                # want no type of caching if just downloading whole thing
-                f.cache = BaseCache(0, f.cache.fetcher, f.size)
-            if getattr(f, "blocksize", 0) and f.size:
-                # opportunity to parallelise here
+        if self.compression:
+            with self.fs._open(path, **kwargs) as f, open(fn, "wb") as f2:
+                if isinstance(f, AbstractBufferedFile):
+                    # want no type of caching if just downloading whole thing
+                    f.cache = BaseCache(0, f.cache.fetcher, f.size)
+                comp = (
+                    infer_compression(path)
+                    if self.compression == "infer"
+                    else self.compression
+                )
+                f = compr[comp](f, mode="rb")
                 data = True
-                block = getattr(f, "blocksize", 5 * 2 ** 20)
                 while data:
+                    block = getattr(f, "blocksize", 5 * 2 ** 20)
                     data = f.read(block)
                     f2.write(data)
-            else:
-                # this only applies to HTTP, should instead use streaming
-                f2.write(f.read())
+        else:
+            self.fs.get(path, fn)
         self.save_cache()
         return self._open(path, mode)
 
@@ -576,7 +649,7 @@ class SimpleCacheFileSystem(WholeFileCacheFileSystem):
 
     def _check_file(self, path):
         self._check_cache()
-        sha = hash_name(path, self.same_names)
+        sha = self.hash_name(path, self.same_names)
         for storage in self.storage:
             fn = os.path.join(storage, sha)
             if os.path.exists(fn):
@@ -591,18 +664,13 @@ class SimpleCacheFileSystem(WholeFileCacheFileSystem):
     def _open(self, path, mode="rb", **kwargs):
         path = self._strip_protocol(path)
 
-        if not path.startswith(self.target_protocol):
-            store_path = self.target_protocol + "://" + path
-        else:
-            store_path = path
-        path = self.fs._strip_protocol(store_path)
         if "r" not in mode:
             return LocalTempFile(self, path, mode=mode)
         fn = self._check_file(path)
         if fn:
             return open(fn, mode)
 
-        sha = hash_name(path, self.same_names)
+        sha = self.hash_name(path, self.same_names)
         fn = os.path.join(self.storage[-1], sha)
         logger.debug("Copying %s to local cache" % path)
         kwargs["mode"] = mode
