@@ -1,8 +1,11 @@
 import requests
+
 from ..spec import AbstractFileSystem
 from ..utils import infer_storage_options
 from .memory import MemoryFile
 
+# TODO: add GIST backend, would be very similar
+
 
 class GithubFileSystem(AbstractFileSystem):
     """Interface to files in github
@@ -25,19 +28,32 @@ class GithubFileSystem(AbstractFileSystem):
     ``sha`` can be the full or abbreviated hex of the commit you want to fetch
     from, or a branch or tag name (so long as it doesn't contain special characters
     like "/", "?", which would have to be HTTP-encoded).
+
+    For authorised access, you must provide username and token, which can be made
+    at https://github.com/settings/tokens
     """
 
     url = "https://api.github.com/repos/{org}/{repo}/git/trees/{sha}"
     rurl = "https://raw.githubusercontent.com/{org}/{repo}/{sha}/{path}"
     protocol = "github"
 
-    def __init__(self, org, repo, sha="master", **kwargs):
+    def __init__(self, org, repo, sha="master", username=None, token=None, **kwargs):
         super().__init__(**kwargs)
         self.org = org
         self.repo = repo
         self.root = sha
+        if (username is None) ^ (token is None):
+            raise ValueError("Auth required both username and token")
+        self.username = username
+        self.token = token
         self.ls("")
 
+    @property
+    def kw(self):
+        if self.username:
+            return {"auth": (self.username, self.token)}
+        return {}
+
     @classmethod
     def repos(cls, org_or_user, is_org=True):
         """List repo names for given org or user
@@ -63,30 +79,77 @@ class GithubFileSystem(AbstractFileSystem):
         r.raise_for_status()
         return [repo["name"] for repo in r.json()]
 
-    def ls(self, path, detail=False, sha=None, **kwargs):
+    @property
+    def tags(self):
+        """Names of tags in the repo"""
+        r = requests.get(
+            "https://api.github.com/repos/{org}/{repo}/tags"
+            "".format(org=self.org, repo=self.repo),
+            **self.kw,
+        )
+        r.raise_for_status()
+        return [t["name"] for t in r.json()]
+
+    @property
+    def branches(self):
+        """Names of branches in the repo"""
+        r = requests.get(
+            "https://api.github.com/repos/{org}/{repo}/branches"
+            "".format(org=self.org, repo=self.repo),
+            **self.kw,
+        )
+        r.raise_for_status()
+        return [t["name"] for t in r.json()]
+
+    @property
+    def refs(self):
+        """Named references, tags and branches"""
+        return {"tags": self.tags, "branches": self.branches}
+
+    def ls(self, path, detail=False, sha=None, _sha=None, **kwargs):
+        """List files at given path
+
+        Parameters
+        ----------
+        path: str
+            Location to list, relative to repo root
+        detail: bool
+            If True, returns list of dicts, one per file; if False, returns
+            list of full filenames only
+        sha: str (optional)
+            List at the given point in the repo history, branch or tag name or commit
+            SHA
+        _sha: str (optional)
+            List this specific tree object (used internally to descend into trees)
+        """
         path = self._strip_protocol(path)
         if path == "":
-            sha = self.root
-        if sha is None:
+            _sha = sha or self.root
+        if _sha is None:
             parts = path.rstrip("/").split("/")
             so_far = ""
-            sha = self.root
+            _sha = sha or self.root
             for part in parts:
-                out = self.ls(so_far, True, sha=sha)
+                out = self.ls(so_far, True, sha=sha, _sha=_sha)
                 so_far += "/" + part if so_far else part
-                out = [o for o in out if o["name"] == so_far][0]
+                out = [o for o in out if o["name"] == so_far]
+                if not out:
+                    raise FileNotFoundError(path)
+                out = out[0]
                 if out["type"] == "file":
                     if detail:
                         return [out]
                     else:
                         return path
-                sha = out["sha"]
-        if path not in self.dircache:
-            r = requests.get(self.url.format(org=self.org, repo=self.repo, sha=sha))
+                _sha = out["sha"]
+        if path not in self.dircache or sha not in [self.root, None]:
+            r = requests.get(
+                self.url.format(org=self.org, repo=self.repo, sha=_sha), **self.kw
+            )
             if r.status_code == 404:
                 raise FileNotFoundError(path)
             r.raise_for_status()
-            self.dircache[path] = [
+            out = [
                 {
                     "name": path + "/" + f["path"] if path else f["path"],
                     "mode": f["mode"],
@@ -96,10 +159,17 @@ class GithubFileSystem(AbstractFileSystem):
                 }
                 for f in r.json()["tree"]
             ]
-        if detail:
-            return self.dircache[path]
+            if sha in [self.root, None]:
+                self.dircache[path] = out
         else:
-            return sorted([f["name"] for f in self.dircache[path]])
+            out = self.dircache[path]
+        if detail:
+            return out
+        else:
+            return sorted([f["name"] for f in out])
+
+    def invalidate_cache(self, path=None):
+        self.dircache.clear()
 
     @classmethod
     def _strip_protocol(cls, path):
@@ -126,14 +196,14 @@ class GithubFileSystem(AbstractFileSystem):
         autocommit=True,
         cache_options=None,
         sha=None,
-        **kwargs
+        **kwargs,
     ):
         if mode != "rb":
             raise NotImplementedError
         url = self.rurl.format(
             org=self.org, repo=self.repo, path=path, sha=sha or self.root
         )
-        r = requests.get(url)
+        r = requests.get(url, **self.kw)
         if r.status_code == 404:
             raise FileNotFoundError(path)
         r.raise_for_status()
