@@ -1,10 +1,11 @@
 import datetime
 import io
 import os
-import shutil
 import posixpath
 import re
+import shutil
 import tempfile
+
 from fsspec import AbstractFileSystem
 from fsspec.utils import stringify_path
 
@@ -22,6 +23,7 @@ class LocalFileSystem(AbstractFileSystem):
 
     root_marker = "/"
     protocol = "file"
+    local_file = True
 
     def __init__(self, auto_mkdir=False, **kwargs):
         super().__init__(**kwargs)
@@ -39,38 +41,53 @@ class LocalFileSystem(AbstractFileSystem):
         os.makedirs(path, exist_ok=exist_ok)
 
     def rmdir(self, path):
+        path = self._strip_protocol(path)
         os.rmdir(path)
 
-    def ls(self, path, detail=False):
+    def ls(self, path, detail=False, **kwargs):
         path = self._strip_protocol(path)
-        paths = [posixpath.join(path, f) for f in os.listdir(path)]
         if detail:
-            return [self.info(f) for f in paths]
+            return [self.info(f) for f in os.scandir(path)]
         else:
-            return paths
+            return [posixpath.join(path, f) for f in os.listdir(path)]
 
-    def glob(self, path, **kargs):
+    def glob(self, path, **kwargs):
         path = self._strip_protocol(path)
-        return super().glob(path)
+        return super().glob(path, **kwargs)
 
     def info(self, path, **kwargs):
-        path = self._strip_protocol(path)
-        out = os.stat(path, follow_symlinks=False)
-        dest = False
-        if os.path.islink(path):
-            t = "link"
-            dest = os.readlink(path)
-        elif os.path.isdir(path):
-            t = "directory"
-        elif os.path.isfile(path):
-            t = "file"
+        if isinstance(path, str):
+            path = self._strip_protocol(path)
+            out = os.stat(path, follow_symlinks=False)
+            link = os.path.islink(path)
+            if os.path.isdir(path):
+                t = "directory"
+            elif os.path.isfile(path):
+                t = "file"
+            else:
+                t = "other"
         else:
-            t = "other"
-        result = {"name": path, "size": out.st_size, "type": t, "created": out.st_ctime}
+            # scandir DirEntry
+            out = path.stat(follow_symlinks=False)
+            link = path.is_symlink()
+            if path.is_dir(follow_symlinks=False):
+                t = "directory"
+            elif path.is_file(follow_symlinks=False):
+                t = "file"
+            else:
+                t = "other"
+            path = self._strip_protocol(path.path)
+        result = {
+            "name": path,
+            "size": out.st_size,
+            "type": t,
+            "created": out.st_ctime,
+            "islink": link,
+        }
         for field in ["mode", "uid", "gid", "mtime"]:
             result[field] = getattr(out, "st_" + field)
-        if dest:
-            result["destination"] = dest
+        if result["islink"]:
+            result["destination"] = os.readlink(path)
             try:
                 out2 = os.stat(path, follow_symlinks=True)
                 result["size"] = out2.st_size
@@ -78,25 +95,31 @@ class LocalFileSystem(AbstractFileSystem):
                 result["size"] = 0
         return result
 
-    def copy(self, path1, path2, **kwargs):
-        shutil.copyfile(path1, path2)
-
-    def get(self, path1, path2, **kwargs):
-        if kwargs.get("recursive"):
-            return super(LocalFileSystem, self).get(path1, path2, **kwargs)
+    def cp_file(self, path1, path2, **kwargs):
+        path1 = self._strip_protocol(path1).rstrip("/")
+        path2 = self._strip_protocol(path2).rstrip("/")
+        if self.auto_mkdir:
+            self.makedirs(self._parent(path2), exist_ok=True)
+        if self.isfile(path1):
+            shutil.copyfile(path1, path2)
+        elif self.isdir(path1):
+            self.mkdirs(path2, exist_ok=True)
         else:
-            return self.copy(path1, path2, **kwargs)
+            raise FileNotFoundError
 
-    def put(self, path1, path2, **kwargs):
-        if kwargs.get("recursive"):
-            return super(LocalFileSystem, self).put(path1, path2, **kwargs)
-        else:
-            return self.copy(path1, path2, **kwargs)
+    def get_file(self, path1, path2, **kwargs):
+        return self.cp_file(path1, path2, **kwargs)
 
-    def mv(self, path1, path2, **kwargs):
-        os.rename(path1, path2)
+    def put_file(self, path1, path2, **kwargs):
+        return self.cp_file(path1, path2, **kwargs)
+
+    def mv_file(self, path1, path2, **kwargs):
+        path1 = self._strip_protocol(path1).rstrip("/")
+        path2 = self._strip_protocol(path2).rstrip("/")
+        shutil.move(path1, path2)
 
     def rm(self, path, recursive=False, maxdepth=None):
+        path = self._strip_protocol(path).rstrip("/")
         if recursive and self.isdir(path):
             shutil.rmtree(path)
         else:
@@ -104,12 +127,14 @@ class LocalFileSystem(AbstractFileSystem):
 
     def _open(self, path, mode="rb", block_size=None, **kwargs):
         path = self._strip_protocol(path)
-        if self.auto_mkdir:
+        if self.auto_mkdir and "w" in mode:
             self.makedirs(self._parent(path), exist_ok=True)
         return LocalFileOpener(path, mode, fs=self, **kwargs)
 
     def touch(self, path, **kwargs):
         path = self._strip_protocol(path)
+        if self.auto_mkdir:
+            self.makedirs(self._parent(path), exist_ok=True)
         if self.exists(path):
             os.utime(path, None)
         else:
@@ -136,7 +161,6 @@ class LocalFileSystem(AbstractFileSystem):
         path = stringify_path(path)
         if path.startswith("file://"):
             path = path[7:]
-        path = os.path.expanduser(path)
         return make_path_posix(path)
 
     def _isfilestore(self):
@@ -148,6 +172,26 @@ class LocalFileSystem(AbstractFileSystem):
 
 def make_path_posix(path, sep=os.sep):
     """ Make path generic """
+    if isinstance(path, (list, set, tuple)):
+        return type(path)(make_path_posix(p) for p in path)
+    if "~" in path:
+        path = os.path.expanduser(path)
+    if sep == "/":
+        # most common fast case for posix
+        if path.startswith("/"):
+            return path
+        return os.getcwd() + "/" + path
+    if (
+        (sep not in path and "/" not in path)
+        or (sep == "/" and not path.startswith("/"))
+        or (sep == "\\" and ":" not in path and not path.startswith("\\\\"))
+    ):
+        # relative path like "path" or "rel\\path" (win) or rel/path"
+        if os.sep == "\\":
+            # abspath made some more '\\' separators
+            return make_path_posix(os.path.abspath(path))
+        else:
+            return os.getcwd() + "/" + path
     if re.match("/[A-Za-z]:", path):
         # for windows file URI like "file:///C:/folder/file"
         # or "file:///C:\\dir\\file"
@@ -162,21 +206,10 @@ def make_path_posix(path, sep=os.sep):
     if path.startswith("\\"):
         # windows network path like "\\server\\path"
         return "/" + path.lstrip("\\").replace("\\", "/").replace("//", "/")
-    if (
-        sep not in path
-        and "/" not in path
-        or (sep == "/" and not path.startswith("/"))
-        or (sep == "\\" and ":" not in path)
-    ):
-        # relative path like "path" or "rel\\path" (win) or rel/path"
-        path = os.path.abspath(path)
-        if os.sep == "\\":
-            # abspath made some more '\\' separators
-            return make_path_posix(path, sep)
     return path
 
 
-class LocalFileOpener(object):
+class LocalFileOpener(io.IOBase):
     def __init__(self, path, mode, autocommit=True, fs=None, **kwargs):
         self.path = path
         self.mode = mode
@@ -197,8 +230,8 @@ class LocalFileOpener(object):
                 self.temp = name
                 self.f = open(name, mode=self.mode)
             if "w" not in self.mode:
-                self.details = self.fs.info(self.path)
-                self.size = self.details["size"]
+                self.size = self.f.seek(0, 2)
+                self.f.seek(0)
                 self.f.size = self.size
 
     def _fetch_range(self, start, end):
@@ -210,13 +243,13 @@ class LocalFileOpener(object):
         return self.f.read(end - start)
 
     def __setstate__(self, state):
+        self.f = None
+        loc = state.pop("loc", None)
+        self.__dict__.update(state)
         if "r" in state["mode"]:
-            loc = self.state.pop("loc")
+            self.f = None
             self._open()
             self.f.seek(loc)
-        else:
-            self.f = None
-        self.__dict__.update(state)
 
     def __getstate__(self):
         d = self.__dict__.copy()
@@ -231,17 +264,51 @@ class LocalFileOpener(object):
     def commit(self):
         if self.autocommit:
             raise RuntimeError("Can only commit if not already set to autocommit")
-        os.replace(self.temp, self.path)
+        shutil.move(self.temp, self.path)
 
     def discard(self):
         if self.autocommit:
             raise RuntimeError("Cannot discard if set to autocommit")
         os.remove(self.temp)
 
+    def readable(self) -> bool:
+        return True
+
+    def writable(self) -> bool:
+        return "r" not in self.mode
+
+    def read(self, *args, **kwargs):
+        return self.f.read(*args, **kwargs)
+
+    def write(self, *args, **kwargs):
+        return self.f.write(*args, **kwargs)
+
+    def tell(self, *args, **kwargs):
+        return self.f.tell(*args, **kwargs)
+
+    def seek(self, *args, **kwargs):
+        return self.f.seek(*args, **kwargs)
+
+    def readline(self, *args, **kwargs):
+        return self.f.readline(*args, **kwargs)
+
+    def readlines(self, *args, **kwargs):
+        return self.f.readlines(*args, **kwargs)
+
+    def close(self):
+        return self.f.close()
+
+    @property
+    def closed(self):
+        return self.f.closed
+
     def __fspath__(self):
         # uniquely among fsspec implementations, this is a real, local path
         return self.path
 
+    def __iter__(self):
+        return self.f.__iter__()
+
     def __getattr__(self, item):
         return getattr(self.f, item)
 
