@@ -1,6 +1,7 @@
+import array
 from collections.abc import MutableMapping
-from .registry import filesystem
-from .core import split_protocol, _un_chain
+
+from .core import url_to_fs
 
 
 class FSMap(MutableMapping):
@@ -19,23 +20,30 @@ class FSMap(MutableMapping):
 
     Examples
     --------
-    >>> fs = FileSystem(**parameters) # doctest: +SKIP
-    >>> d = FSMap('my-data/path/', fs) # doctest: +SKIP
+    >>> fs = FileSystem(**parameters)  # doctest: +SKIP
+    >>> d = FSMap('my-data/path/', fs)  # doctest: +SKIP
     or, more likely
     >>> d = fs.get_mapper('my-data/path/')
 
-    >>> d['loc1'] = b'Hello World' # doctest: +SKIP
-    >>> list(d.keys()) # doctest: +SKIP
+    >>> d['loc1'] = b'Hello World'  # doctest: +SKIP
+    >>> list(d.keys())  # doctest: +SKIP
     ['loc1']
-    >>> d['loc1'] # doctest: +SKIP
+    >>> d['loc1']  # doctest: +SKIP
     b'Hello World'
     """
 
-    def __init__(self, root, fs, check=False, create=False):
+    def __init__(self, root, fs, check=False, create=False, missing_exceptions=None):
         self.fs = fs
         self.root = fs._strip_protocol(root).rstrip(
             "/"
         )  # we join on '/' in _key_to_str
+        if missing_exceptions is None:
+            missing_exceptions = (
+                FileNotFoundError,
+                IsADirectoryError,
+                NotADirectoryError,
+            )
+        self.missing_exceptions = missing_exceptions
         if create:
             if not self.fs.exists(root):
                 self.fs.mkdir(root)
@@ -49,21 +57,70 @@ class FSMap(MutableMapping):
             self.fs.rm(root + "/a")
 
     def clear(self):
-        """Remove all keys below root - empties out mapping
-        """
+        """Remove all keys below root - empties out mapping"""
         try:
             self.fs.rm(self.root, True)
             self.fs.mkdir(self.root)
         except:  # noqa: E722
             pass
 
+    def getitems(self, keys, on_error="raise"):
+        """Fetch multiple items from the store
+
+        If the backend is async-able, this might proceed concurrently
+
+        Parameters
+        ----------
+        keys: list(str)
+            They keys to be fetched
+        on_error : "raise", "omit", "return"
+            If raise, an underlying exception will be raised (converted to KeyError
+            if the type is in self.missing_exceptions); if omit, keys with exception
+            will simply not be included in the output; if "return", all keys are
+            included in the output, but the value will be bytes or an exception
+            instance.
+
+        Returns
+        -------
+        dict(key, bytes|exception)
+        """
+        keys2 = [self._key_to_str(k) for k in keys]
+        oe = on_error if on_error == "raise" else "return"
+        try:
+            out = self.fs.cat(keys2, on_error=oe)
+        except self.missing_exceptions as e:
+            raise KeyError from e
+        out = {
+            k: (KeyError() if isinstance(v, self.missing_exceptions) else v)
+            for k, v in out.items()
+        }
+        return {
+            key: out[k2]
+            for key, k2 in zip(keys, keys2)
+            if on_error == "return" or not isinstance(out[k2], BaseException)
+        }
+
+    def setitems(self, values_dict):
+        """Set the values of multiple items in the store
+
+        Parameters
+        ----------
+        values_dict: dict(str, bytes)
+        """
+        values = {self._key_to_str(k): maybe_convert(v) for k, v in values_dict.items()}
+        self.fs.pipe(values)
+
+    def delitems(self, keys):
+        """Remove multiple keys from the store"""
+        self.fs.rm([self._key_to_str(k) for k in keys])
+
     def _key_to_str(self, key):
         """Generate full path for the key"""
         if isinstance(key, (tuple, list)):
             key = str(tuple(key))
         else:
             key = str(key)
-        return "/".join([self.root, key]) if self.root else key
+        return self.fs._strip_protocol("/".join([self.root, key]) if self.root else key)
 
     def _str_to_key(self, s):
         """Strip path of to leave key name"""
@@ -74,7 +131,7 @@ class FSMap(MutableMapping):
         k = self._key_to_str(key)
         try:
             result = self.fs.cat(k)
-        except (FileNotFoundError, IsADirectoryError):
+        except self.missing_exceptions:
             if default is not None:
                 return default
             raise KeyError(key)
@@ -92,8 +149,7 @@ class FSMap(MutableMapping):
         """Store value in key"""
         key = self._key_to_str(key)
         self.fs.mkdirs(self.fs._parent(key), exist_ok=True)
-        with self.fs.open(key, "wb") as f:
-            f.write(value)
+        self.fs.pipe_file(key, maybe_convert(value))
 
     def __iter__(self):
         return (self._str_to_key(x) for x in self.fs.find(self.root))
@@ -113,18 +169,25 @@ class FSMap(MutableMapping):
         path = self._key_to_str(key)
         return self.fs.exists(path) and self.fs.isfile(path)
 
-    def __getstate__(self):
-        """Mapping should be pickleable"""
-        # TODO: replace with reduce to reinstantiate?
-        return self.fs, self.root
-
-    def __setstate__(self, state):
-        fs, root = state
-        self.fs = fs
-        self.root = root
+    def __reduce__(self):
+        return FSMap, (self.root, self.fs, False, False, self.missing_exceptions)
 
 
-def get_mapper(url, check=False, create=False, **kwargs):
+def maybe_convert(value):
+    if isinstance(value, array.array) or hasattr(value, "__array__"):
+        # bytes-like things
+        value = bytearray(memoryview(value))
+    return value
+
+
+def get_mapper(
+    url,
+    check=False,
+    create=False,
+    missing_exceptions=None,
+    alternate_root=None,
+    **kwargs,
+):
     """Create key-value interface for given URL and options
 
     The URL will be of the form "protocol://location" and point to the root
@@ -143,32 +206,19 @@ def get_mapper(url, check=False, create=False, **kwargs):
     create: bool
         Whether to make the directory corresponding to the root before
         instantiating
+    missing_exceptions: None or tuple
+        If given, these excpetion types will be regarded as missing keys and
+        return KeyError when trying to read data. By default, you get
+        (FileNotFoundError, IsADirectoryError, NotADirectoryError)
+    alternate_root: None or str
+        In cases of complex URLs, the parser may fail to pick the correct part
+        for the mapper root, so this arg can override
 
     Returns
     -------
     ``FSMap`` instance, the dict-like key-value store.
     """
-    chain = _un_chain(url, kwargs)
-    if len(chain) > 1:
-        kwargs = chain[0][2]
-        inkwargs = kwargs
-        urlpath = False
-        for i, ch in enumerate(chain):
-            urls, protocol, kw = ch
-            if isinstance(urls, str):
-                if urlpath is False:
-                    urlpath = protocol + "://" + split_protocol(urls)[1]
-            if i == 0:
-                continue
-            inkwargs["target_protocol"] = protocol
-            inkwargs["target_options"] = kw.copy()
-            inkwargs["fo"] = urls
-            inkwargs = inkwargs["target_options"]
-        protocol = chain[0][1]
-        fs = filesystem(protocol, **kwargs)
-    else:
-        protocol, urlpath = split_protocol(url)
-        fs = filesystem(protocol, **kwargs)
-        urlpath = fs._strip_protocol(url)
     # Removing protocol here - could defer to each open() on the backend
-    return FSMap(urlpath, fs, check, create)
+    fs, urlpath = url_to_fs(url, **kwargs)
+    root = alternate_root if alternate_root is not None else urlpath
+    return FSMap(root, fs, check, create, missing_exceptions=missing_exceptions)
